"""Represent bibliographic information for a single publication.

From the CSL docs:

    Next up are the bibliographic details of the items you wish to cite: the item metadata.

    For example, the bibliographic entry for a journal article may show the names of the
    authors, the year in which the article was published, the article title, the journal
    title, the volume and issue in which the article appeared, the page numbers of the
    article, and the article’s Digital Object Identifier (DOI). All these details help
    the reader identify and find the referenced work.

    Reference managers make it easy to create a library of items. While many reference
    managers have their own way of storing item metadata, most support common bibliographic
    exchange formats such as BibTeX and RIS. The citeproc-js CSL processor introduced a
    JSON-based format for storing item metadata in a way citeproc-js could understand.
    Several other CSL processors have since adopted this “CSL JSON” format (also known as
    “citeproc JSON”).

-- https://github.com/citation-style-language/documentation/blob/master/primer.txt

The terminology we've adopted is csl_data for a list of csl_item dicts, and csl_json
for csl_data that is JSON-serialized.
"""

import copy
import datetime
import logging
import re
from typing import Dict, List, Optional, Union

from manubot.cite.citekey import CiteKey


class CSL_Item(dict):
    """
    CSL_Item represents bibliographic information for a single citeable work.

    On a technical side CSL_Item is a Python dictionary with extra methods
    that help cleaning and manipulating it.

    These methods relate to:
    - adding an `id` key and value for CSL item
    - correcting bibliographic information and its structure
    - adding and reading a custom note to CSL item

    More information on CSL JSON (a list of CSL_Items) is available at:
    - https://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html
    - http://docs.citationstyles.org/en/1.0.1/specification.html#standard-variables
    - https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """

    # The ideas for CSL_Item methods come from the following parts of code:
    #  - [ ] citekey_to_csl_item(citekey, prune=True)
    # The methods in CSL_Item class provide primitives to reconstruct
    # functions above.

    type_mapping = {
        "journal-article": "article-journal",
        "book-chapter": "chapter",
        "posted-content": "manuscript",
        "proceedings-article": "paper-conference",
        "standard": "entry",
        "reference-entry": "entry",
    }

    def __init__(self, dictionary=None, **kwargs) -> None:
        """
        Can use both a dictionary or keywords to create a CSL_Item object:

            CSL_Item(title='The Book')
            CSL_Item({'title': 'The Book'})
            csl_dict = {'title': 'The Book', 'ISBN': '321-321'}
            CSL_Item(csl_dict, type='entry')
            CSL_Item(title='The Book', ISBN='321-321', type='entry')

        CSL_Item object is usually provided by bibliographic information API,
        but constructing small CSL_Item objects is useful for testing.
        """
        if dictionary is None:
            dictionary = {}
        super().__init__(copy.deepcopy(dictionary))
        self.update(copy.deepcopy(kwargs))

    def set_id(self, id_) -> "CSL_Item":
        self["id"] = id_
        return self

    def correct_invalid_type(self) -> "CSL_Item":
        """
        Correct invalid CSL item type.
        Does nothing if `type` not present.

        For detail see https://github.com/CrossRef/rest-api-doc/issues/187
        """
        if "type" in self:
            # Replace a type from in CSL_Item.type_mapping.keys(),
            # leave type intact in other cases.
            t = self["type"]
            self["type"] = self.type_mapping.get(t, t)
        return self

    def set_default_type(self) -> "CSL_Item":
        """
        Set type to 'entry', if type not specified.
        """
        self["type"] = self.get("type", "entry")
        return self

    def prune_against_schema(self) -> "CSL_Item":
        """
        Remove fields that violate the CSL Item JSON Schema.
        """
        from .citeproc import remove_jsonschema_errors

        (csl_item,) = remove_jsonschema_errors([self], in_place=True)
        assert csl_item is self
        return self

    def validate_against_schema(self) -> "CSL_Item":
        """
        Confirm that the CSL_Item validates. If not, raises a
        jsonschema.exceptions.ValidationError.
        """
        from .citeproc import get_jsonschema_csl_validator

        validator = get_jsonschema_csl_validator()
        validator.validate([self])
        return self

    def clean(self, prune: bool = True) -> "CSL_Item":
        """
        Sanitize and touch-up a potentially dirty CSL_Item.
        The following steps are performed:
        - update incorrect values for "type" field when a correct variant is known
        - remove fields that violate the JSON Schema (if prune=True)
        - set default value for "type" if missing, since CSL JSON requires type
        - validate against the CSL JSON schema (if prune=True) to ensure output
          CSL_Item is clean
        """
        logging.debug(
            f"Starting CSL_Item.clean with{'' if prune else 'out'}"
            f"CSL pruning for id: {self.get('id', 'id not specified')}"
        )
        self.correct_invalid_type()
        if prune:
            self.prune_against_schema()
        self.set_default_type()
        if prune:
            self.validate_against_schema()
        return self

    def set_date(
        self,
        date: Union[None, str, datetime.date, datetime.datetime],
        variable: str = "issued",
    ) -> "CSL_Item":
        """
        date: date either as a string (in the form YYYY, YYYY-MM, or YYYY-MM-DD)
            or as a Python date object (datetime.date or datetime.datetime).
        variable: which variable to assign the date to.
        """
        date_parts = date_to_date_parts(date)
        if date_parts:
            self[variable] = {"date-parts": [date_parts]}
        return self

    def get_date(self, variable: str = "issued", fill: bool = False) -> Optional[str]:
        """
        Return a CSL date-variable as ISO formatted string:
        ('YYYY', 'YYYY-MM', 'YYYY-MM-DD', or None).

        variable: which CSL JSON date variable to retrieve
        fill: if True, set missing months to January
            and missing days to the first day of the month.
        """
        try:
            date_parts = self[variable]["date-parts"][0]
        except (IndexError, KeyError):
            return None
        return date_parts_to_string(date_parts, fill=fill)

    @property
    def note(self) -> str:
        """
        Return the value of the "note" field as a string.
        If "note" key is not set, return empty string.
        """
        return str(self.get("note") or "")

    @note.setter
    def note(self, text: str) -> None:
        if text:
            self["note"] = text
        else:
            # if text is None or an empty string, remove the "note" field
            self.pop("note", None)

    @property
    def note_dict(self) -> Dict[str, str]:
        """
        Return a dictionary with key-value pairs encoded by this CSL Item's note.
        Extracts both forms (line-entry and braced-entry) of key-value pairs from the CSL JSON "cheater syntax"
        https://github.com/Juris-M/citeproc-js-docs/blob/93d7991d42b4a96b74b7281f38e168e365847e40/csl-json/markup.rst#cheater-syntax-for-odd-fields

        Assigning to this dict will not update `self["note"]`.
        """
        note = self.note
        line_matches = re.findall(
            r"^(?P<key>[A-Z]+|[-_a-z]+): *(?P<value>.+?) *$", note, re.MULTILINE
        )
        braced_matches = re.findall(
            r"{:(?P<key>[A-Z]+|[-_a-z]+): *(?P<value>.+?) *}", note
        )
        return dict(line_matches + braced_matches)

    def note_append_text(self, text: str) -> None:
        """
        Append text to the note field (as a new line) of a CSL Item.
        If a line already exists equal to `text`, do nothing.
        """
        if not text:
            return
        note = self.note
        if re.search(f"^{re.escape(text)}$", note, flags=re.MULTILINE):
            # do not accumulate duplicate lines of text
            # https://github.com/manubot/manubot/issues/258
            return
        if note and not note.endswith("\n"):
            note += "\n"
        note += text
        self.note = note

    def note_append_dict(self, dictionary: dict) -> None:
        """
        Append key-value pairs specified by `dictionary` to the note field of a CSL Item.
        Uses the the [CSL JSON "cheater syntax"](https://github.com/Juris-M/citeproc-js-docs/blob/93d7991d42b4a96b74b7281f38e168e365847e40/csl-json/markup.rst#cheater-syntax-for-odd-fields)
        to encode additional values not defined by the CSL JSON schema.
        """
        for key, value in dictionary.items():
            if not re.fullmatch(r"[A-Z]+|[-_a-z]+", key):
                logging.warning(
                    f"note_append_dict: skipping adding {key!r} because "
                    f"it does not conform to the variable_name syntax as per https://git.io/fjTzW."
                )
                continue
            if "\n" in value:
                logging.warning(
                    f"note_append_dict: skipping adding {key!r} because "
                    f"the value contains a newline: {value!r}"
                )
                continue
            self.note_append_text(f"{key}: {value}")

    def infer_id(self) -> "CSL_Item":
        """
        Detect and set a non-null/empty value for "id" or else raise a ValueError.
        """
        if self.get("standard_citation"):
            # "standard_citation" field is set with a non-null/empty value
            return self.set_id(self.pop("standard_citation"))
        if self.note_dict.get("standard_id"):
            # "standard_id" note field is set with a non-null/empty value
            return self.set_id(self.note_dict["standard_id"])
        if self.get("id"):
            # "id" field exists and is set with a non-null/empty value
            return self.set_id(self["id"])
        raise ValueError(
            "infer_id could not detect a field with a citation / standard_citation. "
            'Consider setting the CSL Item "id" field.'
        )

    def standardize_id(self) -> "CSL_Item":
        """
        Extract the standard_id (standard citation key) for a csl_item and modify the csl_item in-place to set its "id" field.
        The standard_id is extracted from a "standard_citation" field, the "note" field, or the "id" field.
        The extracted citation is checked for validity and standardized, after which it is the final "standard_id".

        Regarding csl_item modification, the csl_item "id" field is set to the standard_citation and the note field
        is created or updated with key-value pairs for standard_id and original_id.

        Note that the Manubot software generally refers to the "id" of a CSL Item as a citekey.
        However, in this context, we use "id" rather than "citekey" for consistency with CSL's "id" field.
        """
        original_id = self.get("id")
        self.infer_id()
        original_standard_id = self["id"]
        citekey = CiteKey(original_standard_id)
        standard_id = citekey.standard_id
        add_to_note = {}
        note_dict = self.note_dict
        if original_id and original_id != standard_id:
            if original_id != note_dict.get("original_id"):
                add_to_note["original_id"] = original_id
        if original_standard_id and original_standard_id != standard_id:
            if original_standard_id != note_dict.get("original_standard_id"):
                add_to_note["original_standard_id"] = original_standard_id
        if standard_id != note_dict.get("standard_id"):
            add_to_note["standard_id"] = standard_id
        self.note_append_dict(dictionary=add_to_note)
        self.set_id(standard_id)
        return self


def assert_csl_item_type(x) -> None:
    if not isinstance(x, CSL_Item):
        raise TypeError(f"Expected CSL_Item object, got {type(x)}")


def date_to_date_parts(
    date: Union[None, str, datetime.date, datetime.datetime]
) -> Optional[List[int]]:
    """
    Convert a date string or object to a date parts list.

    date: date either as a string (in the form YYYY, YYYY-MM, or YYYY-MM-DD)
        or as a Python date object (datetime.date or datetime.datetime).
    """
    if date is None:
        return None
    if isinstance(date, (datetime.date, datetime.datetime)):
        date = date.isoformat()
    if not isinstance(date, str):
        raise ValueError(f"date_to_date_parts: unsupported type for {date}")
    date = date.strip()
    re_year = r"(?P<year>[0-9]{4})"
    re_month = r"(?P<month>1[0-2]|0[1-9])"
    re_day = r"(?P<day>[0-3][0-9])"
    patterns = [
        f"{re_year}-{re_month}-{re_day}",
        f"{re_year}-{re_month}",
        f"{re_year}",
        ".*",  # regex to match anything
    ]
    for pattern in patterns:
        match = re.match(pattern, date)
        if match:
            break
    date_parts = []
    for part in "year", "month", "day":
        try:
            value = match.group(part)
        except IndexError:
            break
        if not value:
            break
        date_parts.append(int(value))
    if date_parts:
        return date_parts
    return None


def date_parts_to_string(date_parts, fill: bool = False) -> Optional[str]:
    """
    Return a CSL date-parts list as ISO formatted string:
    ('YYYY', 'YYYY-MM', 'YYYY-MM-DD', or None).

    date_parts: list or tuple like [year, month, day] as integers.
        Also supports [year, month] and [year] for situations where the day or month-and-day are missing.
    fill: if True, set missing months to January
        and missing days to the first day of the month.
    """
    if not date_parts:
        return None
    if not isinstance(date_parts, (tuple, list)):
        raise ValueError("date_parts must be a tuple or list")
    while fill and 1 <= len(date_parts) < 3:
        date_parts.append(1)
    widths = 4, 2, 2
    str_parts = []
    for i, part in enumerate(date_parts[:3]):
        width = widths[i]
        if isinstance(part, int):
            part = str(part)
        if not isinstance(part, str):
            break
        part = part.zfill(width)
        if len(part) != width or not part.isdigit():
            break
        str_parts.append(part)
    if not str_parts:
        return None
    iso_str = "-".join(str_parts)
    return iso_str



import json
import logging
import re
from typing import Any, Dict

from .handlers import Handler

CSLItem = Dict[str, Any]


class Handler_URL(Handler):

    standard_prefix = "url"

    prefixes = [
        "url",
        "http",
        "https",
    ]

    def standardize_prefix_accession(self, accession):
        if self.prefix_lower != "url":
            accession = f"{self.prefix_lower}:{accession}"
        return self.standard_prefix, accession

    def get_csl_item(self, citekey):
        return get_url_csl_item(citekey.standard_accession)


def get_url_csl_item(url: str) -> CSLItem:
    """
    Get csl_item for a URL trying a sequence of strategies.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `url_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    for retriever in url_retrievers:
        try:
            return retriever(url)
        except Exception as error:
            logging.warning(
                f"Error in {retriever.__name__} for {url} "
                f"due to a {error.__class__.__name__}:\n{error}"
            )
            logging.info(error, exc_info=True)
    raise Exception(f"all get_url_csl_item methods failed for {url}")


def get_url_csl_item_zotero(url: str) -> CSLItem:
    """
    Use Zotero's translation-server to generate a CSL Item for the specified URL.
    """
    from manubot.cite.zotero import export_as_csl, web_query

    zotero_data = web_query(url)
    csl_data = export_as_csl(zotero_data)
    (csl_item,) = csl_data
    if not csl_item.get("URL"):
        # some Zotero translators don't set URL. https://github.com/manubot/manubot/issues/244
        csl_item["URL"] = url
    return csl_item


def get_url_csl_item_greycite(url: str) -> CSLItem:
    """
    Uses Greycite which has experiened uptime problems in the past.
    API calls seem to take at least 15 seconds. Browser requests are much
    faster. Setting header did not have an effect. Consider mimicking browser
    using selenium.

    More information on Greycite at:
    http://greycite.knowledgeblog.org/
    http://knowledgeblog.org/greycite
    https://arxiv.org/abs/1304.7151
    https://git.io/v9N2C
    """
    import requests

    from manubot.util import get_manubot_user_agent

    headers = {
        "Connection": "close",  # https://github.com/kennethreitz/requests/issues/4023
        "User-Agent": get_manubot_user_agent(),
    }
    response = requests.get(
        "http://greycite.knowledgeblog.org/json", params={"uri": url}, headers=headers
    )
    response.raise_for_status()
    # Some Greycite responses were valid JSON besides for an error appended
    # like "<p>*** Date set from uri<p>" or "<p>*** fetch error : 404<p>".
    pattern = re.compile(r"<p>\*\*\*.*<p>")
    text = pattern.sub("", response.text)
    csl_item = json.loads(text)
    csl_item["type"] = "webpage"
    return csl_item


def get_url_csl_item_manual(url: str) -> CSLItem:
    """
    Manually create csl_item for a URL.
    """
    return {"URL": url, "type": "webpage"}


url_retrievers = [
    get_url_csl_item_zotero,
    get_url_csl_item_greycite,
    get_url_csl_item_manual,
]



import json
import logging
import os
import random
import re
import warnings
from datetime import date
from typing import List, Optional

import jinja2

from manubot.process.ci import get_continuous_integration_parameters
from manubot.process.manuscript import (
    datetime_now,
    get_manuscript_stats,
    get_text,
)
from manubot.process.metadata import (
    get_head_commit,
    get_header_includes,
    get_manuscript_urls,
    get_software_versions,
    get_thumbnail_url,
)
from manubot.util import (
    get_configured_yaml,
    read_serialized_data,
    read_serialized_dict,
)


def read_variable_files(paths: List[str], variables: Optional[dict] = None) -> dict:
    """
    Read multiple serialized data files into a user_variables dictionary.
    Provide `paths` (a list of URLs or local file paths).
    Paths can optionally have a namespace prepended.
    For example:

    ```python
    paths = [
        'https://git.io/vbkqm',  # update the dictionary's top-level
        'namespace_1=https://git.io/vbkqm',  # store under 'namespace_1' key
        'namespace_2=some_local_path.json',  # store under 'namespace_2' key
    ]
    ```

    If a namespace is not provided, the JSON must contain a dictionary as its
    top level. Namespaces should consist only of ASCII alphanumeric characters
    (includes underscores, first character cannot be numeric).

    Pass a dictionary to `variables` to update an existing dictionary rather
    than create a new dictionary.
    """
    if variables is None:
        variables = {}
    for path in paths:
        logging.info(f"Reading user-provided templating variables at {path!r}")
        # Match only namespaces that are valid jinja2 variable names
        # http://jinja.pocoo.org/docs/2.10/api/#identifier-naming
        match = re.match(r"([a-zA-Z_][a-zA-Z0-9_]*)=(.+)", path)
        if match:
            namespace, path = match.groups()
            logging.info(
                f"Using the {namespace!r} namespace for template variables from {path!r}"
            )
        try:
            if match:
                obj = {namespace: read_serialized_data(path)}
            else:
                obj = read_serialized_dict(path)
        except Exception:
            logging.exception(f"Error reading template variables from {path!r}")
            continue
        assert isinstance(obj, dict)
        conflicts = variables.keys() & obj.keys()
        if conflicts:
            logging.warning(
                f"Template variables in {path!r} overwrite existing "
                "values for the following keys:\n" + "\n".join(conflicts)
            )
        variables.update(obj)
    logging.debug(
        f"Reading user-provided templating variables complete:\n"
        f"{json.dumps(variables, indent=2, ensure_ascii=False)}"
    )
    return variables


def _convert_field_to_list(
    dictionary, field, separator=False, deprecation_warning_key=None
):
    """
    Convert `dictionary[field]` to a list. If value is a string and
    `separator` is specified, split by `separator`. If `deprecation_warning_key`
    is provided, warn when `dictionary[field]` is a string.
    """
    if field not in dictionary:
        return dictionary
    value = dictionary[field]
    if isinstance(value, list):
        return dictionary
    if isinstance(value, str):
        if separator is False:
            dictionary[field] = [value]
        else:
            dictionary[field] = value.split(separator)
        if deprecation_warning_key:
            warnings.warn(
                f"Expected list for {dictionary.get(deprecation_warning_key)}'s {field}. "
                + (
                    f"Assuming multiple {field} are `{separator}` separated. "
                    if separator
                    else ""
                )
                + f"Please switch {field} to a list.",
                category=DeprecationWarning,
            )
        return dictionary
    raise ValueError("Unsupported value type {value.__class__.__name__}")


def randomize_authors(authors: list, seed=None) -> list:
    """
    Randomize author list based on a given random seed.
    See https://github.com/manubot/manubot/issues/315.
    """
    return random.Random(seed).sample(authors, k=len(authors))


def add_author_affiliations(variables: dict) -> dict:
    """
    Edit variables to contain numbered author affiliations. Specifically,
    add a list of affiliation_numbers for each author and add a list of
    affiliations to the top-level of variables. If no authors have any
    affiliations, variables is left unmodified.
    """
    affiliations = []
    for author in variables["authors"]:
        _convert_field_to_list(
            dictionary=author,
            field="affiliations",
            separator="; ",
            deprecation_warning_key="name",
        )
        _convert_field_to_list(
            dictionary=author,
            field="funders",
        )
        affiliations.extend(author.get("affiliations", []))
    if not affiliations:
        return variables
    affiliations = list(dict.fromkeys(affiliations))  # deduplicate
    affil_to_number = {affil: i for i, affil in enumerate(affiliations, start=1)}
    for author in variables["authors"]:
        numbers = [affil_to_number[affil] for affil in author.get("affiliations", [])]
        author["affiliation_numbers"] = sorted(numbers)
    variables["affiliations"] = [
        {"affiliation": affil, "affiliation_number": i}
        for affil, i in affil_to_number.items()
    ]
    return variables


def load_variables(args) -> dict:
    """
    Read `metadata.yaml` and files specified by `--template-variables-path` to generate
    manuscript variables available for jinja2 templating.

    Returns a dictionary, refered to as `variables`, with the following keys:

    - `pandoc`: a dictionary for passing options to Pandoc via the `yaml_metadata_block`.
      Fields in `pandoc` are either generated by Manubot or hard-coded by the user if `metadata.yaml`
      includes a `pandoc` dictionary.
    - `manubot`: a dictionary for manubot-related information and metadata.
      Fields in `manubot` are either generated by Manubot or hard-coded by the user if `metadata.yaml`
      includes a `manubot` dictionary.
    - All fields from a manuscript's `metadata.yaml` that are not interpreted by Manubot are
      copied to `variables`. Interpreted fields include `pandoc`, `manubot`, `title`,
      `keywords`, `authors` (formerly `author_info`, now deprecated), `lang`, and `thumbnail`.
    - User-specified fields inserted according to the `--template-variables-path` option.
      User-specified variables take highest precedence and can overwrite values for existing
      keys like `pandoc` or `manubot` (dangerous).
    """
    # Generated manuscript variables
    variables = {"pandoc": {}, "manubot": {}}

    # Read metadata which contains pandoc_yaml_metadata
    # as well as authors information.
    if args.meta_yaml_path.is_file():
        metadata = read_serialized_dict(args.meta_yaml_path)
    else:
        metadata = {}
        logging.warning(
            f"missing {args.meta_yaml_path} file with yaml_metadata_block for pandoc"
        )

    # Interpreted keys that are intended for pandoc
    move_to_pandoc = "title", "keywords", "lang"
    for key in move_to_pandoc:
        if key in metadata:
            variables["pandoc"][key] = metadata.pop(key)

    # Add date & generated timestamp to metadata
    manuscript_date = metadata.pop("date", None)
    if isinstance(manuscript_date, str):
        manuscript_date = date.fromisoformat(manuscript_date)
    now = datetime_now()
    if not manuscript_date:
        manuscript_date = now.date()
        logging.info(
            "No explicit manuscript date provided. "
            f"Dating manuscript based on the current datetime: {now.isoformat()} "
            f"(in the {now:%Z} timezone)"
        )
    variables["pandoc"]["date-meta"] = manuscript_date.isoformat()
    variables["manubot"]["date"] = manuscript_date.isoformat()
    variables["manubot"][
        "date_long"
    ] = f"{manuscript_date:%B} {manuscript_date.day}, {manuscript_date.year}"
    variables["manubot"]["generated"] = now.isoformat(timespec="seconds")
    variables["manubot"]["generated_date_long"] = f"{now:%B} {now.day}, {now.year}"

    # Process authors metadata
    if "author_info" in metadata:
        authors = metadata.pop("author_info", [])
        warnings.warn(
            "metadata.yaml: 'author_info' is deprecated. Use 'authors' instead.",
            category=DeprecationWarning,
        )
    else:
        authors = metadata.pop("authors", [])
    if authors is None:
        authors = []
    # author order randomization
    variables["manubot"]["randomize_author_order"] = bool(
        metadata.pop("manubot-randomize-author-order", False)
    )
    if variables["manubot"]["randomize_author_order"]:
        variables["manubot"]["randomize_author_order_seed"] = (
            metadata.pop("manubot-randomize-author-order-seed", None)
            or get_head_commit()
        )
        authors = randomize_authors(
            authors, seed=variables["manubot"]["randomize_author_order_seed"]
        )
    # set author variables
    variables["pandoc"]["author-meta"] = [author["name"] for author in authors]
    variables["manubot"]["authors"] = authors
    add_author_affiliations(variables["manubot"])

    # Set repository version metadata for CI builds
    ci_params = get_continuous_integration_parameters()
    if ci_params:
        variables["manubot"]["ci_source"] = ci_params

    # Add manuscript URLs
    variables["manubot"].update(get_manuscript_urls(metadata.pop("html_url", None)))

    # Add software versions
    variables["manubot"].update(get_software_versions(rootstock=not args.skip_remote))

    # Add thumbnail URL if present
    thumbnail_url = get_thumbnail_url(metadata.pop("thumbnail", None))
    if thumbnail_url:
        variables["manubot"]["thumbnail_url"] = thumbnail_url

    # Update variables with metadata.yaml pandoc/manubot dicts
    for key in "pandoc", "manubot":
        dict_ = metadata.pop(key, {})
        if not isinstance(dict_, dict):
            logging.warning(
                f"load_variables expected metadata.yaml field {key!r} to be a dict."
                f"Received a {dict_.__class__.__name__!r} instead."
            )
            continue
        variables[key].update(dict_)

    # Update variables with uninterpreted metadata.yaml fields
    variables.update(metadata)

    # Update variables with user-provided variables here
    variables = read_variable_files(args.template_variables_path, variables)

    # Add header-includes metadata with <meta> information for the HTML output's <head>
    variables["pandoc"]["header-includes"] = get_header_includes(variables)

    assert args.skip_citations
    # Extend Pandoc's metadata.bibliography field with manual references paths
    bibliographies = variables["pandoc"].get("bibliography", [])
    if isinstance(bibliographies, str):
        bibliographies = [bibliographies]
    assert isinstance(bibliographies, list)
    bibliographies.extend(args.manual_references_paths)
    bibliographies = list(map(os.fspath, bibliographies))
    variables["pandoc"]["bibliography"] = bibliographies
    # enable pandoc-manubot-cite option to write bibliography to a file
    variables["pandoc"]["manubot-output-bibliography"] = os.fspath(args.references_path)
    variables["pandoc"]["manubot-output-citekeys"] = os.fspath(args.citations_path)
    variables["pandoc"]["manubot-requests-cache-path"] = os.fspath(
        args.requests_cache_path
    )
    variables["pandoc"]["manubot-clear-requests-cache"] = args.clear_requests_cache

    return variables


def template_with_jinja2(text, variables):
    """
    Template using jinja2 with the variables dictionary unpacked as keyword
    arguments.
    """
    jinja_environment = jinja2.Environment(
        loader=jinja2.BaseLoader(),
        undefined=jinja2.make_logging_undefined(logging.getLogger()),
        autoescape=False,
        comment_start_string="{##",
        comment_end_string="##}",
        extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        # A trailing newline is critical to prevent pyyaml
        # from adding a block chomping strip indicator to
        # header-includes metadata, which caused an unwanted Pandoc behavior.
        # https://github.com/jgm/pandoc/issues/8449
        keep_trailing_newline=True,
    )
    template = jinja_environment.from_string(text)
    return template.render(**variables)


def prepare_manuscript(args):
    """
    Compile manuscript, creating manuscript.md and references.json as inputs
    for pandoc.
    """
    text = get_text(args.content_directory)
    assert args.skip_citations

    variables = load_variables(args)
    variables["manubot"]["manuscript_stats"] = get_manuscript_stats(text)
    with args.variables_path.open("w", encoding="utf-8") as write_file:
        json.dump(variables, write_file, ensure_ascii=False, indent=2)
        write_file.write("\n")

    text = template_with_jinja2(text, variables)

    # Write manuscript for pandoc
    yaml = get_configured_yaml()
    with args.manuscript_path.open("w", encoding="utf-8") as write_file:
        yaml.dump(
            variables["pandoc"],
            write_file,
            default_flow_style=False,
            explicit_start=True,
            explicit_end=True,
            width=float("inf"),
            allow_unicode=True,
            sort_keys=False,
        )
        write_file.write("\n")
        write_file.write(text)



"""
Utilities for representing and processing citation keys.
"""
import dataclasses
import functools
import logging
import re
import typing as tp

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property


@dataclasses.dataclass
class CiteKey:
    input_id: str
    """Input identifier for the citekey"""
    aliases: dict = dataclasses.field(default_factory=dict)
    """Mapping from input identifier to aliases"""
    infer_prefix: bool = True
    """Whether to infer the citekey's prefix when a prefix is missing or unhandled"""

    def __post_init__(self):
        self.check_input_id(self.input_id)

    @staticmethod
    def check_input_id(input_id) -> None:
        if not isinstance(input_id, str):
            raise TypeError(
                "input_id should be type 'str' not "
                f"{type(input_id).__name__!r}: {input_id!r}"
            )
        if input_id.startswith("@"):
            raise ValueError(f"invalid citekey input_id: {input_id!r}\nstarts with '@'")

    @classmethod
    @functools.lru_cache(maxsize=None)
    def from_input_id(cls, *args, **kwargs) -> "CiteKey":
        """Cached constructor"""
        return cls(*args, **kwargs)

    @cached_property
    def dealiased_id(self) -> str:
        """
        If `self.input_id` is in `self.aliases`, the value specified by
        `self.aliases`. Otherwise, `self.input_id`.
        """
        return self.aliases.get(self.input_id, self.input_id)

    def _set_prefix_accession(self) -> None:
        self._prefix = None
        self._accession = None
        split_id = self.dealiased_id.split(":", 1)
        if len(split_id) == 2:
            self._prefix, self._accession = split_id
        if self.infer_prefix and not self.is_known_prefix:
            self._infer_prefix()

    def _infer_prefix(self) -> None:
        """
        Treat `self.dealiased_id` as missing a prefix.
        If the prefix can be inferred, set `self._prefix` and `self._accession`.

        Only call this function from _set_prefix_accession,
        since it is not safe after instance attributes or properties have been cached.
        """
        from .handlers import infer_prefix

        prefix = infer_prefix(self.dealiased_id)
        if not prefix:
            return
        self._prefix = prefix
        self._accession = self.dealiased_id

    @property
    def prefix(self) -> tp.Optional[str]:
        """
        If `self.input_id` contains a colon, the substring up to the first colon.
        Otherwise, None.
        """
        if not hasattr(self, "_prefix"):
            self._set_prefix_accession()
        return self._prefix

    @property
    def prefix_lower(self) -> tp.Optional[str]:
        """
        A lowercase version of `self.prefix` or None.
        """
        if self.prefix is None:
            return None
        return self.prefix.lower()

    @property
    def accession(self) -> tp.Optional[str]:
        """
        If `self.prefix`, the remainder of `self.input_id` following the first colon.
        """
        if not hasattr(self, "_accession"):
            self._set_prefix_accession()
        return self._accession

    @property
    def standard_prefix(self) -> tp.Optional[str]:
        """
        If the citekey is handled, the standard prefix specified by the handler.
        Otherwise, None.
        """
        if not hasattr(self, "_standard_prefix"):
            self._standardize()
        return self._standard_prefix

    @property
    def standard_accession(self) -> tp.Optional[str]:
        """
        If the citekey is handled, the standard accession specified by the handler.
        Otherwise, None.
        """
        if not hasattr(self, "_standard_accession"):
            self._standardize()
        return self._standard_accession

    @cached_property
    def handler(self):
        from .handlers import Handler, get_handler

        if self.is_handled_prefix:
            return get_handler(self.prefix_lower)
        return Handler(self.prefix_lower)

    @property
    def is_handled_prefix(self) -> bool:
        from .handlers import prefix_to_handler

        return self.prefix_lower in prefix_to_handler

    @property
    def is_known_prefix(self) -> bool:
        return self.is_handled_prefix or self.is_pandoc_xnos_prefix()

    def inspect(self) -> tp.Optional[str]:
        """
        Inspect citekey for potential problems.
        If no problems are found, return None.
        Otherwise, returns a string describing the problem.
        """
        return self.handler.inspect(self)

    def _standardize(self) -> None:
        """
        Set `self._standard_prefix`, `self._standard_accession`, and `self._standard_id`.
        For citekeys without a prefix or with an unhandled prefix, _standard_prefix
        and _standard_accession are set to None.
        """
        if not self.is_handled_prefix:
            self._standard_prefix = None
            self._standard_accession = None
            self._standard_id = self.dealiased_id
            return
        fxn = self.handler.standardize_prefix_accession
        self._standard_prefix, self._standard_accession = fxn(self.accession)
        self._standard_id = f"{self._standard_prefix}:{self._standard_accession}"

    @property
    def standard_id(self) -> str:
        """
        If the citekey is handled, the standard_id specified by the handler.
        Otherwise, `self.dealiased_id`.
        """
        if not hasattr(self, "_standard_id"):
            self._standardize()
        return self._standard_id

    @cached_property
    def short_id(self) -> str:
        """
        A hashed version of standard_id whose characters are
        within the ranges 0-9, a-z and A-Z.
        """
        return shorten_citekey(self.standard_id)

    @cached_property
    def all_ids(self) -> tp.List[str]:
        ids = [self.input_id, self.dealiased_id, self.standard_id, self.short_id]
        ids = [x for x in ids if x]  # remove None
        ids = list(dict.fromkeys(ids))  # deduplicate
        return ids

    def __hash__(self):
        return hash((self.input_id, self.dealiased_id))

    def __repr__(self):
        return " --> ".join(
            f"{getattr(self, key)} ({key})"
            for key in (
                "input_id",
                "dealiased_id",
                "prefix_lower",
                "accession",
                "standard_id",
                "short_id",
            )
        )

    @cached_property
    def csl_item(self):
        from .csl_item import CSL_Item

        csl_item = self.handler.get_csl_item(self)
        if not isinstance(csl_item, CSL_Item):
            csl_item = CSL_Item(csl_item)
        csl_item.set_id(self.standard_id)
        return csl_item

    def is_pandoc_xnos_prefix(self, log_case_warning: bool = False) -> bool:
        from .handlers import _pandoc_xnos_prefixes

        if self.prefix in _pandoc_xnos_prefixes:
            return True
        if log_case_warning and self.prefix_lower in _pandoc_xnos_prefixes:
            logging.warning(
                "pandoc-xnos prefixes should be all lowercase.\n"
                f'Should {self.input_id!r} use {self.prefix_lower!r} rather than "{self.prefix!r}"?'
            )
        return False


def shorten_citekey(standard_citekey: str) -> str:
    """
    Return a shortened citekey derived from the input citekey.
    The input citekey should be standardized prior to this function,
    since differences in the input citekey will result in different shortened citekeys.
    Short citekeys are generated by converting the input citekey to a 6 byte hash,
    and then converting this digest to a base62 ASCII str. Shortened
    citekeys consist of characters in the following ranges: 0-9, a-z and A-Z.
    """
    import hashlib

    import base62

    assert not standard_citekey.startswith("@")
    as_bytes = standard_citekey.encode()
    blake_hash = hashlib.blake2b(as_bytes, digest_size=6)
    digest = blake_hash.digest()
    short_citekey = base62.encodebytes(digest)
    return short_citekey


def citekey_to_csl_item(
    citekey, prune=True, manual_refs=None, log_level: tp.Union[str, int] = "WARNING"
):
    """
    Generate a CSL_Item for the input citekey.
    """
    from manubot import __version__ as manubot_version

    if manual_refs is None:
        manual_refs = {}

    # https://stackoverflow.com/a/35704430/4651668
    log_level = logging._checkLevel(log_level)
    if not isinstance(citekey, CiteKey):
        citekey = CiteKey(citekey)

    if citekey.standard_id in manual_refs:
        return manual_refs[citekey.standard_id]

    try:
        csl_item = citekey.csl_item
    except Exception as error:
        logging.log(
            log_level,
            f"Generating csl_item for {citekey.standard_id!r} failed "
            f"due to a {error.__class__.__name__}:\n{error}",
        )
        logging.info(error, exc_info=True)
        return None
    # update csl_item with manubot generated metadata
    note_text = f"This CSL Item was generated by Manubot v{manubot_version} from its persistent identifier (standard_id)."
    note_dict = {"standard_id": citekey.standard_id}
    csl_item.note_append_text(note_text)
    csl_item.note_append_dict(note_dict)
    csl_item.set_id(citekey.short_id)
    csl_item.clean(prune=prune)
    return csl_item


def url_to_citekey(url: str) -> str:
    """
    Convert a HTTP(s) URL into a citekey.
    For supported sources, convert from url citekey to an alternative source like doi.
    If citekeys fail inspection, revert alternative sources to URLs.
    """
    from urllib.parse import unquote, urlparse

    citekey = None
    parsed_url = urlparse(url)
    domain_levels = parsed_url.hostname.split(".")
    if domain_levels[-2:] == ["doi", "org"]:
        # DOI URLs
        doi = unquote(parsed_url.path.lstrip("/"))
        citekey = f"doi:{doi}"
    if domain_levels[-2] == "sci-hub":
        # Sci-Hub domains
        doi = parsed_url.path.lstrip("/")
        citekey = f"doi:{doi}"
    if domain_levels[-2:] == ["biorxiv", "org"]:
        # bioRxiv URL to DOI. See https://git.io/Je9Hq
        match = re.search(
            r"/(?P<biorxiv_id>([0-9]{4}\.[0-9]{2}\.[0-9]{2}\.)?[0-9]{6,})",
            parsed_url.path,
        )
        if match:
            citekey = f"doi:10.1101/{match.group('biorxiv_id')}"
    is_ncbi_url = parsed_url.hostname.endswith("ncbi.nlm.nih.gov")
    if is_ncbi_url and parsed_url.path.startswith("/pubmed/"):
        # PubMed URLs
        try:
            pmid = parsed_url.path.split("/")[2]
            citekey = f"pmid:{pmid}"
        except IndexError:
            pass
    if is_ncbi_url and parsed_url.path.startswith("/pmc/"):
        # PubMed Central URLs
        try:
            pmcid = parsed_url.path.split("/")[3]
            citekey = f"pmcid:{pmcid}"
        except IndexError:
            pass
    if domain_levels[-2:] == ["wikidata", "org"] and parsed_url.path.startswith(
        "/wiki/"
    ):
        # Wikidata URLs
        try:
            wikidata_id = parsed_url.path.split("/")[2]
            citekey = f"wikidata:{wikidata_id}"
        except IndexError:
            pass
    if domain_levels[-2:] == ["arxiv", "org"]:
        # arXiv identifiers. See https://arxiv.org/help/arxiv_identifier
        try:
            arxiv_id = parsed_url.path.split("/", maxsplit=2)[2]
            if arxiv_id.endswith(".pdf"):
                arxiv_id = arxiv_id[:-4]
            citekey = f"arxiv:{arxiv_id}"
        except IndexError:
            pass
    if citekey is None or CiteKey(citekey).inspect() is not None:
        citekey = f"url:{url}"
    return citekey



import json
import logging
import os
import subprocess
from typing import Any, Dict, List, Optional

from manubot.pandoc.util import get_pandoc_info
from manubot.util import shlex_join


def load_bibliography(
    path: Optional[str] = None,
    text: Optional[str] = None,
    input_format: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Convert a bibliography to CSL JSON using either `pandoc-citeproc --bib2json`
    or `pandoc --to=csljson`, depending on availability of pandoc commands on the system.
    Accepts either a bibliography path or text (string). If supplying text,
    pandoc-citeproc will likely require input_format be specified.
    The CSL JSON is returned as Python objects.
    If loading fails, log an error and return an empty list.

    Parameters
    ----------
    path : str, pathlike, or None
        Path to a bibliography file. Extension is used by pandoc-citeproc to infer the
        format of the input.
    text : str or None
        Text representation of the bibliography, such as a JSON-formatted string.
        `input_format` should be specified if providing text input.
    input_format : str or None
        Manually specified input formatted that is supported by pandoc-citeproc:
        https://github.com/jgm/pandoc-citeproc/blob/master/man/pandoc-citeproc.1.md#options
        Use 'bib' for BibLaTeX. Use 'json' for CSL JSON.

    Returns
    -------
    csl_json : JSON-like object
        CSL JSON Data for the references encoded by the input bibliography.
    """
    use_text = path is None
    use_path = text is None
    if use_path:
        path = os.fspath(path)
    if not (use_text ^ use_path):
        raise ValueError("load_bibliography: specify either path or text but not both.")
    pdoc_info = get_pandoc_info()
    if pdoc_info["pandoc-citeproc"]:
        return _load_bibliography_pandoc_citeproc(path, text, input_format)
    if input_format == "bib" or (use_path and path.endswith(".bib")):
        return _load_bibliography_pandoc(path, text)
    logging.error(
        "pandoc-citeproc not found on system, but is required to convert any format besides 'bib': "
        "manubot.pandoc.bibliography.load_bibliography returning empty CSL JSON"
    )
    return []


def _load_bibliography_pandoc_citeproc(
    path: Optional[str] = None,
    text: Optional[str] = None,
    input_format: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Convert a bibliography to CSL JSON using `pandoc-citeproc --bib2json`.
    Accepts either a bibliography path or text (string). If supplying text,
    pandoc-citeproc will likely require input_format be specified.
    The CSL JSON is returned as Python objects.
    If loading fails, log an error and return an empty list.

    Parameters
    ----------
    path : str, pathlike, or None
        Path to a bibliography file. Extension is used by pandoc-citeproc to infer the
        format of the input.
    text : str or None
        Text representation of the bibliography, such as a JSON-formatted string.
        `input_format` should be specified if providing text input.
    input_format : str or None
        Manually specified input formatted that is supported by pandoc-citeproc:
        https://github.com/jgm/pandoc-citeproc/blob/master/man/pandoc-citeproc.1.md#options

    Returns
    -------
    csl_json : JSON-like object
        CSL JSON Data for the references encoded by the input bibliography.
    """
    command_args = ["pandoc-citeproc", "--bib2json"]
    if input_format:
        command_args.extend(["--format", input_format])
    return _pandoc_system_call(command_args, path, text)


def _load_bibliography_pandoc(
    path: Optional[str] = None,
    text: Optional[str] = None,
) -> List[Dict[str, Any]]:

    """
    Convert a biblatex (.bib) bibliography to CSL JSON data using pandoc directly.
    Pandoc support for csljson output requires pandoc >= 2.11.
    """
    pdoc_info = get_pandoc_info()
    if not pdoc_info["pandoc"]:
        logging.error(
            "pandoc not found on system: "
            "manubot.pandoc.bibliography.load_bibliography returning empty CSL JSON"
        )
        return []
    if pdoc_info["pandoc version"] < (2, 11):
        logging.error(
            "pandoc >= version 2.11 required for biblatex to csljson conversion. "
            "manubot.pandoc.bibliography.load_bibliography returning empty CSL JSON"
        )
        return []
    command_args = "pandoc --from=biblatex --to=csljson".split()
    return _pandoc_system_call(command_args, path, text)


def _pandoc_system_call(
    command_args: List[str], path: Optional[str], text: Optional[str]
) -> List[Dict[str, Any]]:
    """
    Call "pandoc citeproc" or "pandoc" using input from a path or text.
    Return dict representing CSL JSON.
    """
    assert command_args[0].startswith("pandoc")
    run_kwargs = {}
    if path:
        command_args.append(os.fspath(path))
    else:
        run_kwargs["input"] = text
    logging.info("load_bibliography subprocess args:\n>>> " + shlex_join(command_args))
    process = subprocess.run(
        command_args,
        capture_output=True,
        encoding="utf-8",
        **run_kwargs,
    )
    logging.info(f"captured stderr:\n{process.stderr}")
    if process.returncode:
        logging.error(
            f"Pandoc call returned nonzero exit code.\n"
            f"{shlex_join(process.args)}\n{process.stderr}"
        )
        return []
    try:
        csl_json = json.loads(process.stdout)
    except (TypeError, json.decoder.JSONDecodeError):
        logging.error(f"Error parsing bib2json output as JSON:\n{process.stdout}")
        csl_json = []
    return csl_json



"""
This module defines a pandoc filter for manubot cite functionality.
The filter can be called with the `pandoc-manubot-cite` command.

## Options

Configuration is provided via Pandoc metadata fields.

- `bibliography` (sequence of strings):
  Use to define reference metadata manually.
  Pandoc supports specifying multiple external bibliography files.
  When bibliography files are specified,
  this filter will read them instead of pandoc.
  Behavior should be similar to Pandoc,
  with format inferred by extension:
  .json for CSL JSON,
  .yaml for CSL YAML,
  .bib for BibLaTeX.

- `references` (sequence of CSL-Item mappings):
  Same as Pandoc's references metadata field.

- `citekey-aliases` (mapping: string -> string):
  Used to define aliases (tags) for cite-by-id citations.
  Useful when a citation is used many times or contains invalid characters.
  Aliases can also be defined in markdown with link reference syntax.

- `manubot-infer-citekey-prefixes` (boolean):
  Attempt to infer the prefix for citekeys without a known prefix.
  For example, allow '@10.1371/journal.pcbi.1007128' with a 'doi:' prefix.
  Default is true.

- `manubot-output-citekeys` (string):
  path to write TSV table of citekeys
  showing their transformation from input_id to short_id.

- `manubot-bibliography-cache` (string):
  Path to read and write bibliographic metadata as CSL JSON/YAML.
  Intended as a human-editable cache of the bibliography data,
  for situations where this filter is run multiple times.
  This is similar to specifying bibliography=FILE and manubot-output-bibliography=FILE in a single argument,
  but will not error trying to read a bibliography file that does not yet exist.

- `manubot-output-bibliography` (string):
  path to write generated CSL JSON bibliography.
  If specified in addition to `manubot-bibliography-cache`,
  two output bibliographies will be written (with the same references).

- `manubot-requests-cache-path` (string):
  Enable caching HTTP requests to this path (minus the extension) using [requests-cache](https://github.com/reclosedev/requests-cache).
  For example, setting to `.cache/requests-cache` will cache requests to `.cache/requests-cache.sqlite`.

- `manubot-clear-requests-cache` (boolean):
  If true, clear the requests cache at `manubot-requests-cache-path`.

- `manubot-fail-on-errors` (boolean):
  If true, return a nonzero exit status if any errors are logged.
  Default is false, which allows Pandoc to proceed when some citations could not be processed.

## development commands

```shell
# export to plain text (with pandoc < 2.11)
pandoc \
  --to=plain \
  --standalone \
  --filter=pandoc-manubot-cite \
  --filter=pandoc-citeproc \
  manubot/pandoc/tests/test_cite_filter/input.md

# call the filter manually using pandoc JSON output
pandoc \
  --to=json \
  manubot/pandoc/tests/test_cite_filter/input.md \
  | python manubot/pandoc/test_cite.py markdown
```

Related resources on pandoc filters:

- [python pandocfilters package](https://github.com/jgm/pandocfilters)
- [python panflute package](https://github.com/sergiocorreia/panflute)
- [panflute Citation class](http://scorreia.com/software/panflute/code.html#panflute.elements.Citation)
"""
import argparse
import logging
import os
from typing import Any, Dict

import panflute as pf

from manubot.cite.citations import Citations


def parse_args() -> argparse.Namespace:
    """
    Read command line arguments
    """
    parser = argparse.ArgumentParser(
        description="Pandoc filter for citation by persistent identifier. "
        "Filters are command-line programs that read and write a JSON-encoded abstract syntax tree for Pandoc. "
        "Unless you are debugging, run this filter as part of a pandoc command by specifying --filter=pandoc-manubot-cite."
    )
    parser.add_argument(
        "target_format",
        help="output format of the pandoc command, as per Pandoc's --to option",
    )
    parser.add_argument(
        "--input",
        nargs="?",
        type=argparse.FileType("r", encoding="utf-8"),
        help="path read JSON input (defaults to stdin)",
    )
    parser.add_argument(
        "--output",
        nargs="?",
        type=argparse.FileType("w", encoding="utf-8"),
        help="path to write JSON output (defaults to stdout)",
    )
    args = parser.parse_args()
    return args


def _get_citekeys_action(elem: pf.Element, doc: pf.Doc) -> None:
    """
    Panflute action to extract citationId from all Citations in the AST.
    """
    if not isinstance(elem, pf.Citation):
        return None
    manuscript_citekeys = doc.manubot["manuscript_citekeys"]
    manuscript_citekeys.append(elem.id)
    return None


def _citation_to_id_action(elem: pf.Element, doc: pf.Doc) -> None:
    """
    Panflute action to update the citationId of Citations in the AST
    with their manubot-created keys.
    """
    if not isinstance(elem, pf.Citation):
        return None
    mapper = doc.manubot["citekey_shortener"]
    if elem.id in mapper:
        elem.id = mapper[elem.id]
    return None


def _get_reference_link_citekey_aliases(elem: pf.Element, doc: pf.Doc) -> None:
    """
    Extract citekey aliases from the document that were defined
    using markdown's link reference syntax.
    https://spec.commonmark.org/0.29/#link-reference-definitions

    Based on pandoc-url2cite implementation by phiresky at
    https://github.com/phiresky/pandoc-url2cite/blob/b28374a9a037a5ce1747b8567160d8dffd64177e/index.ts#L118-L152
    """
    if type(elem) != pf.Para:
        # require link reference definitions to be in their own paragraph
        return
    while (
        len(elem.content) >= 3
        and type(elem.content[0]) == pf.Cite
        and len(elem.content[0].citations) == 1
        and type(elem.content[1]) == pf.Str
        and elem.content[1].text == ":"
    ):
        # paragraph consists of at least a Cite (with one Citation),
        # a Str (equal to ":"), and additional elements, such as a
        # link destination and possibly more link-reference definitions.
        dest_index = 3 if type(elem.content[2]) in {pf.Space, pf.SoftBreak} else 2
        destination = elem.content[dest_index]
        if type(destination) != pf.Str:
            # prevent infinite loop as per https://github.com/manubot/manubot/pull/302#issuecomment-906743300
            logging.warning(
                f"Unsupported destination type {type(destination)} in link reference syntax:\n{elem.to_json()}"
            )
            break
        # paragraph starts with `[@something]: something`
        # save info to citekeys and remove from paragraph
        citekey = elem.content[0].citations[0].id
        citekey_aliases = doc.manubot["citekey_aliases"]
        if citekey in citekey_aliases and citekey_aliases[citekey] != destination.text:
            logging.warning(f"multiple aliases defined for @{citekey}")
        citekey_aliases[citekey] = destination.text
        # found citation, add it to citekeys and remove it from document
        elem.content = elem.content[dest_index + 1 :]
        # remove leading SoftBreak, before continuing
        if len(elem.content) > 0 and type(elem.content[0]) == pf.SoftBreak:
            elem.content.pop(0)


def _get_load_manual_references_kwargs(doc: pf.Doc) -> Dict[str, Any]:
    """
    Return keyword arguments for Citations.load_manual_references.
    """
    manual_refs = doc.get_metadata("references", default=[])
    bibliography_paths = doc.get_metadata("bibliography", default=[])
    if not isinstance(bibliography_paths, list):
        bibliography_paths = [bibliography_paths]
    bibliography_cache_path = doc.manubot["bibliography_cache"]
    if (
        bibliography_cache_path
        and bibliography_cache_path not in bibliography_paths
        and os.path.exists(bibliography_cache_path)
    ):
        bibliography_paths.append(bibliography_cache_path)
    return {
        "paths": bibliography_paths,
        "extra_csl_items": manual_refs,
    }


def process_citations(doc: pf.Doc) -> None:
    """
    Apply citation-by-identifier to a Python object representation of
    Pandoc's Abstract Syntax Tree.
    """
    # process metadata.manubot-bibliography-cache
    bib_cache = doc.get_metadata(key="manubot-bibliography-cache")
    if not (bib_cache is None or isinstance(bib_cache, str)):
        logging.warning(
            f"Expected metadata.manubot-bibliography-cache to be a string or null (None), "
            f"but received a {bib_cache.__class__.__name__}. Setting to None."
        )
        bib_cache = None
    doc.manubot["bibliography_cache"] = bib_cache
    # process metadata.citekey-aliases
    citekey_aliases = doc.get_metadata("citekey-aliases", default={})
    if not isinstance(citekey_aliases, dict):
        logging.warning(
            f"Expected metadata.citekey-aliases to be a dict, "
            f"but received a {citekey_aliases.__class__.__name__}. Disregarding."
        )
        citekey_aliases = {}
    doc.manubot["citekey_aliases"] = citekey_aliases
    doc.walk(_get_reference_link_citekey_aliases)
    doc.walk(_get_citekeys_action)
    manuscript_citekeys = doc.manubot["manuscript_citekeys"]
    citations = Citations(
        input_ids=manuscript_citekeys,
        aliases=citekey_aliases,
        infer_citekey_prefixes=doc.get_metadata(
            "manubot-infer-citekey-prefixes", default=True
        ),
    )
    citations.csl_item_failure_log_level = "ERROR"

    requests_cache_path = doc.get_metadata("manubot-requests-cache-path")
    if requests_cache_path:
        from manubot.process.requests_cache import RequestsCache

        req_cache = RequestsCache(requests_cache_path)
        req_cache.mkdir()
        req_cache.install()
        if doc.get_metadata("manubot-clear-requests-cache", default=False):
            req_cache.clear()

    citations.filter_pandoc_xnos()
    citations.load_manual_references(**_get_load_manual_references_kwargs(doc))
    citations.inspect(log_level="WARNING")
    citations.get_csl_items()
    doc.manubot["citekey_shortener"] = citations.input_to_csl_id
    doc.walk(_citation_to_id_action)

    if requests_cache_path:
        req_cache.close()

    citations.write_citekeys_tsv(path=doc.get_metadata("manubot-output-citekeys"))
    citations.write_csl_items(path=doc.get_metadata("manubot-output-bibliography"))
    citations.write_csl_items(path=doc.manubot["bibliography_cache"])
    # Update pandoc metadata with fields that this filter
    # has either consumed, created, or modified.
    doc.metadata["bibliography"] = []
    doc.metadata["references"] = citations.csl_items
    doc.metadata["citekey_aliases"] = citekey_aliases


def main() -> None:
    from manubot.command import (
        exit_if_error_handler_fired,
        setup_logging_and_errors,
    )

    diagnostics = setup_logging_and_errors()
    args = parse_args()
    # Let panflute handle io to sys.stdout / sys.stdin to set utf-8 encoding.
    # args.input=None for stdin, args.output=None for stdout
    doc = pf.load(input_stream=args.input)
    log_level = doc.get_metadata("manubot-log-level", "WARNING")
    diagnostics["logger"].setLevel(getattr(logging, log_level))
    logging.debug(f"Input Pandoc metadata:\n{doc.get_metadata()}")
    doc.manubot = {"manuscript_citekeys": []}
    process_citations(doc)
    pf.dump(doc, output_stream=args.output)
    if doc.get_metadata("manubot-fail-on-errors", False):
        exit_if_error_handler_fired(diagnostics["error_handler"])


if __name__ == "__main__":
    main()



import argparse
import json
import logging
import pathlib
import subprocess
import sys

from manubot.cite.citations import Citations
from manubot.pandoc.util import get_pandoc_info
from manubot.util import shlex_join

# For manubot cite, infer --format from --output filename extensions
extension_to_format = {
    ".json": "csljson",
    ".yaml": "cslyaml",
    ".yml": "cslyaml",
    ".txt": "plain",
    ".md": "markdown",
    ".docx": "docx",
    ".html": "html",
    ".xml": "jats",
}


def call_pandoc(metadata, path, format="plain"):
    """
    path is the path to write to.
    """
    _exit_without_pandoc()
    info = get_pandoc_info()
    _check_pandoc_version(info, metadata, format)
    metadata_block = "---\n{yaml}\n...\n".format(
        yaml=json.dumps(metadata, ensure_ascii=False, indent=2)
    )
    args = [
        "pandoc",
        "--citeproc"
        if info["pandoc version"] >= (2, 11)
        else "--filter=pandoc-citeproc",
        f"--output={path or '-'}",
    ]
    if format == "markdown":
        args.extend(["--to=markdown_strict-raw_html", "--wrap=none"])
    elif format == "jats":
        args.extend(["--to=jats", "--standalone"])
    elif format == "docx":
        args.extend(["--to=docx"])
    elif format == "html":
        args.extend(["--to=html"])
    elif format == "plain":
        args.extend(["--to=plain", "--wrap=none"])
        if info["pandoc version"] >= (2,):
            # Do not use ALL_CAPS for bold & underscores for italics
            # https://github.com/jgm/pandoc/issues/4834#issuecomment-412972008
            filter_path = (
                pathlib.Path(__file__)
                .joinpath("..", "plain-pandoc-filter.lua")
                .resolve()
            )
            assert filter_path.exists()
            args.append(f"--lua-filter={filter_path}")
    logging.info("call_pandoc subprocess args:\n" + shlex_join(args))
    process = subprocess.run(
        args=args,
        input=metadata_block.encode(),
    )
    process.check_returncode()


def _parse_cli_cite_args(args: argparse.Namespace):
    arg_dict = vars(args)
    # infer format from output extension
    if not args.format and args.output:
        arg_dict["format"] = extension_to_format.get(args.output.suffix)
    # default format to csljson
    if not args.format:
        arg_dict["format"] = "csljson"
    # whether to render references with Pandoc
    arg_dict["render"] = args.format not in {"csljson", "cslyaml"}
    logging.debug(f"_parse_cli_cite_args: {args}")


def cli_cite(args: argparse.Namespace):
    """
    Main function for the manubot cite command-line interface.

    Does not allow user to directly specify Pandoc's --to argument, due to
    inconsistent citation rendering by output format. See
    https://github.com/jgm/pandoc/issues/4834
    """
    _parse_cli_cite_args(args)
    citations = Citations(
        input_ids=args.citekeys,
        infer_citekey_prefixes=args.infer_prefix,
        prune_csl_items=args.prune_csl,
        sort_csl_items=False,
    )
    citations.load_manual_references(paths=args.bibliography)
    citations.inspect(log_level="WARNING")
    csl_items = citations.get_csl_items()

    # output CSL data, if --render is False
    if not args.render:
        if args.format == "csljson":
            text = citations.csl_json
        elif args.format == "cslyaml":
            text = citations.csl_yaml
        else:
            raise ValueError("format must be csljson or cslyaml")
        write_file = args.output.open("wb") if args.output else sys.stdout.buffer
        with write_file:
            write_file.write(text.encode())
        return

    # use Pandoc to render references
    pandoc_metadata = {"nocite": "@*", "csl": args.csl, "references": csl_items}
    call_pandoc(metadata=pandoc_metadata, path=args.output, format=args.format)


def _exit_without_pandoc() -> None:
    """
    Given info from get_pandoc_info, exit Python if Pandoc is not available.
    """
    if get_pandoc_info()["pandoc"]:
        return
    logging.critical(
        "pandoc command not found on system. Ensure that Pandoc is installed."
    )
    raise SystemExit(1)


def _check_pandoc_version(info, metadata, format):
    """
    Given info from get_pandoc_info, check that Pandoc's version is sufficient
    to perform the citation rendering command specified by metadata and format.
    Please add additional minimum version information to this function, as its
    discovered.
    """
    issues = []
    if format == "jats" and info["pandoc version"] < (2,):
        issues.append("--jats requires pandoc >= v2.0.")
    # --csl=URL did not work in https://travis-ci.org/greenelab/manubot/builds/417314743#L796,
    # but exact version where this fails unknown
    # if metadata.get('csl', '').startswith('http') and pandoc_version < (2,):
    #     issues.append('--csl=URL requires pandoc >= v2.0.')
    issues = "\n".join(issues)
    if issues:
        logging.critical(f"issues with pandoc version detected:\n{issues}")



import functools
import logging
import shutil
import subprocess
from typing import Any, Dict, Tuple


@functools.lru_cache()
def get_pandoc_info() -> Dict[str, Any]:
    """
    Return path and version information for the system's pandoc and
    pandoc-citeproc commands. When Pandoc is installed,
    the output will look like:
    ```python
    {
        'pandoc': True,
        'pandoc path': '/PATH_TO_EXECUTABLES/pandoc',
        'pandoc version': (2, 5),
        'pandoc-citeproc': True,
        'pandoc-citeproc path': '/PATH_TO_EXECUTABLES/pandoc-citeproc',
        'pandoc-citeproc version': (0, 15),
    }
    ```

    If the executables are missing, the output will be like:
    ```python
    {
        'pandoc': False,
        'pandoc-citeproc': False,
    }
    ```
    """

    stats = get_command_info("pandoc")
    if stats["pandoc version"] < (2, 11):
        stats.update(get_command_info("pandoc-citeproc"))
    else:
        stats["pandoc-citeproc"] = False

    logging.info("\n".join(f"{k}: {v}" for k, v in stats.items()))
    return stats


def get_pandoc_version() -> Tuple[int, ...]:
    """
    Return pandoc version as tuple of major and minor
    version numbers, for example: (2, 7, 2)
    """
    pandoc_info = get_pandoc_info()
    if not pandoc_info["pandoc"]:
        # https://twitter.com/dhimmel/status/1327082301994496000
        raise ImportError("missing pandoc command on system.")
    return pandoc_info["pandoc version"]


def get_command_info(command: str) -> dict:
    """
    Returns a dictionary containing some information about a command
    """

    command_info_dict = {}

    path = shutil.which(command)
    command_info_dict[command] = bool(path)

    if not path:
        return command_info_dict

    version = subprocess.check_output(args=[command, "--version"], encoding="utf-8")
    logging.debug(version)
    version, *_discard = version.splitlines()
    _discard, version = version.strip().split()
    from packaging.version import parse as parse_version

    version = parse_version(version).release
    command_info_dict[f"{command} version"] = version
    command_info_dict[f"{command} path"] = path

    return command_info_dict
